@@ -1,14 +1,14 @@
-/* @(#)update.c	1.123 12/12/22 Copyright 1985, 88, 91, 1995-2012 J. Schilling */
+/* @(#)update.c	1.126 14/03/25 Copyright 1985, 88, 91, 1995-2014 J. Schilling */
 #include <schily/mconfig.h>
 #ifndef lint
 static	UConst char sccsid[] =
-	"@(#)update.c	1.123 12/12/22 Copyright 1985, 88, 91, 1995-2012 J. Schilling";
+	"@(#)update.c	1.126 14/03/25 Copyright 1985, 88, 91, 1995-2014 J. Schilling";
 #endif
 /*
  *	Make program
  *	Macro handling / Dependency Update
  *
- *	Copyright (c) 1985, 88, 91, 1995-2012 by J. Schilling
+ *	Copyright (c) 1985, 88, 91, 1995-2014 by J. Schilling
  */
 /*
  * The contents of this file are subject to the terms of the
@@ -17,6 +17,8 @@ static	UConst char sccsid[] =
  * with the License.
  *
  * See the file CDDL.Schily.txt in this distribution for details.
+ * A copy of the CDDL is also available via the Internet at
+ * http://www.opensource.org/licenses/cddl1.txt
  *
  * When distributing Covered Code, include this CDDL HEADER in each
  * file and include the License file CDDL.Schily.txt from this distribution.
@@ -31,6 +33,7 @@ static	UConst char sccsid[] =
 #include <schily/schily.h>
 #include <schily/libport.h>
 #include "make.h"
+#include "job.h"
 
 EXPORT obj_t	*default_tgt;		/* Current 'make' arg or default tgt */
 EXPORT BOOL	found_make;		/* Did we expand the $(MAKE) macro?  */
@@ -554,28 +557,28 @@ build_path(level, name, namelen, path, psize)
 	size_t	psize;
 {
 	list_t *lp;
-	char	*dirname = (char *)NULL;
+	char	*dir_name = (char *)NULL;
 	register int n = level;
 
 	if (n <= 1) {
 		if (level == OBJLEVEL) {
-			dirname = ObjDir;
+			dir_name = ObjDir;
 			namelen += slashlen + ObjDirlen;
 		}
 	} else if (level != MAXLEVEL) {
 		if ((lp = list_nth(SearchList, n - 2)) == (list_t *)NULL)
 			return ((char *)NULL);
-		dirname = lp->l_obj->o_name;
+		dir_name = lp->l_obj->o_name;
 		namelen += slashlen + lp->l_obj->o_namelen;
 	}
-	if (dirname == (char *)NULL)
+	if (dir_name == (char *)NULL)
 		return (name);
 
 	if (namelen >= psize) {
 		psize = namelen + 1;
 		path = ___realloc(NULL, psize, "build path name");
 	}
-	n = snprintf(path, psize, "%s%s%s", dirname, slash, name);
+	n = snprintf(path, psize, "%s%s%s", dir_name, slash, name);
 	if (n >= psize)
 		etoolong("build path name", name);
 
@@ -1124,9 +1127,19 @@ exp_var(end, cmd, obj, source, suffix, depends)
 	register int	nestlevel = 0;
 		BOOL	funccall = FALSE;
 
-/*error("end: %c cmd: %.50s\n", end, cmd);*/
+/*error("exp_var(end: '%c' cmd: '%.50s') [beg: '%c']\n", end, cmd, beg);*/
 	pat[0] = '\0';
-	while ((ch = *s) != '\0' && ch != ':' && !white(ch)) {
+	while ((ch = *s) != '\0') {
+		if (nestlevel == 0) {
+			/*
+			 * Only stop looking for the end of the macro name in
+			 * case we are at the initial nestlevel. Collect as
+			 * much as possible for a recursive macro name
+			 * expansion below.
+			 */
+			if (ch == ':' || white(ch))
+				break;
+		}
 		if (ch == beg)
 			nestlevel++;
 		if (ch == end)
@@ -1163,7 +1176,7 @@ exp_var(end, cmd, obj, source, suffix, depends)
 /*		return (cmd);*/
 	}
 
-	if (*s == ':' || *s == ' ') {
+	if (*s == ':' || white(*s)) {
 		rname = pat;
 		if (funccall) {
 			while (*s && white(*s))
@@ -1171,6 +1184,7 @@ exp_var(end, cmd, obj, source, suffix, depends)
 		} else {
 			s++;
 		}
+		nestlevel = 0;			/* Paranoia */
 		while ((ch = *s) != '\0') {
 			if (ch == beg)
 				nestlevel++;
